/*
 * Copyright (c) 2013, 
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Nenad Corbic <ncorbic@sangoma.com>
 * 		 Pushkar Singh <psingh@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#include "sample_tapping.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int exit_sample_app(int max_wait);
int map_tapping_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t *span_conf);

int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj);

/* Application LOGGING MODULE FUNCTION Declaration*/
ftdm_status_t app_open_log_file(void);
void app_logger(const char *file, const char *func, int line, int level, const char *fmt, ...);

span_data_t* sample_app_get_ftdm_span(uint32_t span_id);

ftdm_status_t app_wp_config_input();

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1
#define MAX_FILE_PATH_SIZE 25

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;
int g_init = 0;
FILE *log_file_ptr;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };
sangoma_freetdm_init_t sample_app_ftdm_init;
sng_wp_config_t g_wp_config;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : on_sigusr()
 *
 * Desc  : In case if there is any iterruption by user in order then stop the 
 * 	   application and exit
 *
 * Ret   : void
**********************************************************************************/
#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;

	case SIGUSR2:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;
	}

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	getchar();
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}

/**********************************************************************************
 * Fun   : on_sigint()
 *
 * Desc  : In case if SIGINT is received then stop the application and exit
 *
 * Ret   : void
**********************************************************************************/
void on_sigint(int signal)
{
	g_running = 0;

	ftdm_log(FTDM_LOG_INFO, "Signal[%d] recieved ...\n", signal);

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");
	getchar();

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}
#endif

/****************************************
 FreeTDM Event Callbacks
****************************************/
/**********************************************************************************
 * Fun   : sample_app_hwprobe_callback()
 *
 * Desc  : Stop all the ports one by one then save the configuration and then start
 * 	   the respective wanpipe/port. Wait for some time in order to get the
 * 	   wanpipe in connected state and create respecteive FreeTDM spans
 *
 * Ret   : FTDM_SUCCESS=0 | FTDM_FAIL | -1
**********************************************************************************/
int sample_app_hwprobe_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int err;

	/* Filter out all other cards just look for cards Tapping is supported */
	switch (card_type) {
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
	case AFT_ADPTR_ISDN:
		break;

	default:
		ftdm_log(FTDM_LOG_INFO, "%s card (%d) does not support ISDN Tapping\n", SNG_ISDN_TAP_CARD_TYPE(card_type),
				 card_type);
		return FTDM_FAIL;
	}
#if 0
	/* If we want to skip some ports then use below condition to return from here..
	 * Sample app by-def will start all the system present wanpipe ports */
	if (port_id != 1 && port_id != 2) {
		return FTDM_FAIL;
	}
#endif

	SPANS[port_id].init = 1;
	SPANS[port_id].sangoma_port_obj = sangoma_port_obj;

	/* Stop the respective port/wanpipe */
	err = sangoma_config_port_stop(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
	}

	/*********************************************************************/
	/* Below is only for sample app..copying global wp data to each port */
	memcpy(&SPANS[port_id].wp_conf, &g_wp_config, sizeof(sng_wp_config_t));
	/*********************************************************************/

	/* Save Application defined configuration */
	if (SPANS[port_id].wp_conf.wp_config_pres)
		sangoma_set_wp_config(sangoma_port_obj, &SPANS[port_id].wp_conf);

	/* Save the respective port/wanpipe configuration */
	err = sangoma_config_port_save(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
		return -1;
	}

	/* Start the respective port/wanpipe */
	err = sangoma_config_port_start(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to start sangoma port config\n");
		return -1;
	}

	/* Create spans in FreeTDM */
	err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdm_span, SNG_ISDN_PRI_TAPPING_MODE);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to create freetdm span %i\n", port_id);
		return -1;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : sample_app_get_ftdm_span()
 *
 * Desc  : This function will go through with configured ports and get the
 * 	    application port configured structure based on freetdm span id
 *
 * Ret   : span_data_t | NULL
**********************************************************************************/
span_data_t *sample_app_get_ftdm_span(uint32_t span_id)
{
	int i=0;
	for (i = 1;i<FTDM_MAX_SPANS_INTERFACE;i++) {
		if (!SPANS[i].ftdm_span) continue;
		if (span_id == ftdm_span_get_id(SPANS[i].ftdm_span)) {
			return &SPANS[i];
		}
	}
	return NULL;
}

/**********************************************************************************
 * Fun   : sample_app_exec_io_stats()
 *
 * Desc  : Set default ip status
 *
 * Ret   : void
**********************************************************************************/
void sample_app_exec_io_stats(const char* cmd)
{
	int i=0;
	if (!cmd) 
		return;
	for (i = 1;i<FTDM_MAX_SPANS_INTERFACE;i++) {
		if (!SPANS[i].ftdm_span) continue;
		sangoma_freetdm_exec_io_stats(cmd, SPANS[i].ftdm_span);
	}
}

/**********************************************************************************
 * Fun   : sample_app_set_rx_queuesize()
 *
 * Desc  : Set span signalling channel RX queue size
 *
 * Ret   : void
**********************************************************************************/
void sample_app_set_rx_queuesize(int rxqueue_size)
{
	int i=0;
	for (i = 1;i<FTDM_MAX_SPANS_INTERFACE;i++) {
		if (!SPANS[i].ftdm_span) continue;
		sangoma_freetdm_set_span_sig_queuesize(SPANS[i].ftdm_span, rxqueue_size); 
	}
}

/**********************************************************************************
 * Fun   : sample_app_span_start_callback()
 *
 * Desc  : For Tapping one needs to carry config and start span once all spans
 * 	   gets created as for tapping purpose one have to provide the peerspan
 * 	   for each span configured. This function mainly configures spans and 
 * 	   start the same.
 *
 * Ret   : FTDM_SUCCESS | -1 | ret
**********************************************************************************/
int sample_app_span_start_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int ret = 0;

	/* Filter out all other cards just look for cards Tapping is supported */
	ret = map_tapping_spans(sangoma_port_obj, port_id, &SPANS[port_id].span_conf);
	if (ret != FTDM_SUCCESS) {
		return ret;
	}
	
	/* Loads the configuration with respect to TAPPING module */
	if (sangoma_freetdm_span_config_and_start(sangoma_port_obj, SPANS[port_id].ftdm_span, &SPANS[port_id].span_conf, SNG_ISDN_PRI_TAPPING_MODE)) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to load and configure freetdm span %i\n", port_id);
		return -1;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : tapping_main_menu()
 *
 * Desc  : Provides option in order to stop the application at any point of time
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
int tapping_main_menu(void)
{
	char choice[6];

	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "	    SANGOMA PRI/BRI TAPPING MODULE STARTED             \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Please enter your choice:  \n");
	fprintf(stdout, "1: Calls Statistics \n");
	fprintf(stdout, "2: Exit \n");
	fprintf(stdout, "Please enter 1 at any point of time if you want to exit from application \n");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			return FTDM_FAIL;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Exit from application
		 */
		switch (choice[0]) {
		case '1':
			sample_app_exec_io_stats("stats");
			sangoma_freetdm_isdn_tap_statistics();
			
			break;

		case '2':
			fprintf(stdout, "SANGOMA Tapping Sample Application is getting closed...\n");

			/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
			 * the configired wanpipes/hwports */
			exit_sample_app(20);

			fprintf(stdout, "\n\n\n=========================================================\n");
			fprintf(stdout, "                      GOOD BYE                            \n");
			fprintf(stdout, "=========================================================\n");

			return FTDM_SUCCESS;

		default:
			break;
		}
	}

	return FTDM_FAIL;
}

/**********************************************************************************
 * Fun   : start_tapping()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which 
 * 	   call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *start_tapping(ftdm_thread_t *me, void *obj)
{
	uint8_t frame[1024];
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	ftdm_channel_t *chan = NULL;
	ftdm_span_t *span = NULL;
	ftdm_size_t len;
	uint32_t span_id = sigmsg->span_id;
	uint32_t chan_id = sigmsg->chan_id;
	ftdm_caller_data_t *caller_data = NULL;
	char timestamp[80];
	char path[512];
	ftdm_size_t dlen=0x00;
	int fd = -1;

	memset(path, 0 , sizeof(path));
	memset(timestamp, 0 , sizeof(timestamp));

	ftdm_log(FTDM_LOG_DEBUG, "Enter: start_tapping on span_id[%d] chanid[%d] \n", span_id, chan_id);

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		goto done;
	}

	if (!span_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Span ID\n");
		goto done;
	}

	ftdm_span_find(span_id, &span);

	/* Send request to Freetdm to start tracing the media */
	if (!chan_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel ID\n");
		goto done;
	}

	if (!span) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid span found\n");
		goto done;
	}
	chan = ftdm_span_get_channel(span, chan_id);

	if (!chan) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel found\n");
		goto done;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Start tapping on span %d and channel %d\n", span_id, chan_id);

	sangoma_get_timestamp_str(timestamp);

	caller_data = ftdm_channel_get_caller_data(chan);

	if (caller_data) {
		if (!ftdm_strlen_zero(caller_data->dnis.digits)) {
			sprintf(path, "recordings/%s-%d-%d-%s", timestamp, span_id, chan_id, caller_data->dnis.digits);
		} else {
			sprintf(path, "recordings/%s-%d-%d", timestamp, span_id, chan_id);
		}
	} else {
		sprintf(path, "recordings/%s-%d-%d", timestamp, span_id, chan_id);
	}

	if (!strlen(path)) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Path\n");
		goto done;
	}

	if ((fd = open_audio_file(chan_id, span_id, path)) == -1) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to start audio tracing on span[%d]-chan[%d]\n", span_id, chan_id);
		goto done;
	}

	/* Start reading the media, and write to a file in order to tap the call */ 
	while (ftdm_channel_call_check_answered(chan)) {
		ftdm_wait_flag_t flags = FTDM_READ;

		if (ftdm_channel_wait(chan, &flags, -1) == FTDM_FAIL) {
			break;
		}
		len = sizeof(frame);

		if (flags & FTDM_READ) {
			if (ftdm_channel_read(chan, frame, &len) == FTDM_SUCCESS) {
				/* Passive Tapping application should not write to channel, only read and write the data
				 * to a file */
				dlen = len;
				if ((ftdm_size_t)write(fd, frame, (int)dlen) != dlen) {
					ftdm_log(FTDM_LOG_WARNING, "Raw input trace failed to write all of the %"FTDM_SIZE_FMT" bytes\n", dlen);
				}
			}
		}

		/* No data to read..continue with checking..as if call terminate this loop will automatically end due to while condition */
	}

	if (fd >= 0) {
		ftdm_log(FTDM_LOG_DEBUG, "Stop Tapping and closing [%s] file with fd[%d], as call is hanged upon span[%d] chan[%d]\n", path, fd, span_id, chan_id);
		close(fd);
		fd = -1;
	}

  done:
	ftdm_safe_free(sigmsg->raw.data);
	ftdm_safe_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : tap_handle_raw_message()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which
 * 	   call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *tap_handle_raw_message(ftdm_thread_t *me, void *obj)
{
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	isdn_tap_msg_info_t *tap_msg = NULL;		/* tap raw message */
	uint32_t span_id = 0;				/* span id on which call is present */
	uint32_t chan_id = 0;				/* chan id on which call is present */
	uint32_t call_ref = 0;				/* call rference value of the call */
	tap_call_direction_t call_dir;			/* call direction i.e. incoming or outgoing */
	uint8_t call_con = 0;				/* to detect whether call is connected or not */
	char clg_num[25];				/* Calling party number if present */
	char cld_num[25];				/* Called party number if present */

	memset(cld_num, 0, sizeof(cld_num));
	memset(clg_num, 0, sizeof(clg_num));

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		return NULL;
	}

	if (!(sigmsg->raw.data)) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid raw data present in signalling message\n");
		goto done;
	}

	tap_msg = sigmsg->raw.data;

	span_id = tap_msg->span_id;
	chan_id = tap_msg->chan_id;
	call_ref = tap_msg->call_ref;
	call_dir = tap_msg->call_dir;
	call_con = tap_msg->call_con;

	ftdm_log(FTDM_LOG_INFO, "*****************************************\n");
	ftdm_log(FTDM_LOG_INFO, "              Call Details               \n");
	ftdm_log(FTDM_LOG_INFO, "*****************************************\n");

	if (span_id) {
		ftdm_log(FTDM_LOG_INFO, "Call Direction : %s\n", ISDN_CALL_DIRECTION(call_dir));
	}

	ftdm_log(FTDM_LOG_INFO, "Call Received of Span ID : %d\n", span_id);
	ftdm_log(FTDM_LOG_INFO, "Call Received of Chan ID : %d\n", chan_id);
	ftdm_log(FTDM_LOG_INFO, "Call reference value : %d\n", call_ref);

	if (call_con) {
		ftdm_log(FTDM_LOG_INFO, "Call Type : Successfull Call\n");
	} else {
		ftdm_log(FTDM_LOG_INFO, "Call Type : Missed Call\n");
	}

	if (!ftdm_strlen_zero(tap_msg->calling_num)) {
		ftdm_set_string(clg_num, tap_msg->calling_num);
		ftdm_log(FTDM_LOG_INFO, "Calling Number : %s\n", clg_num);
	}

	if (!ftdm_strlen_zero(tap_msg->called_num)) {
		ftdm_set_string(cld_num, tap_msg->called_num);
		ftdm_log(FTDM_LOG_INFO, "Called Number : %s\n", cld_num);
	}

	ftdm_log(FTDM_LOG_INFO, "*****************************************\n\n");

	ftdm_safe_free(sigmsg->raw.data);
  done:
	ftdm_safe_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : app_freetdm_event_queue_handler()
 *
 * Desc  : Starts respective io operation w.r.t different event received 
 *
 * Ret   : NULL
**********************************************************************************/
static void *app_freetdm_event_queue_handler(ftdm_thread_t *me, void *obj)
{
	ftdm_interrupt_t 	*ftdm_sangoma_isdn_int[1];
	ftdm_status_t 		ret_status;
	int32_t 		sleep = SNGISDN_TAP_EVENT_POLL_RATE;
	ftdm_sigmsg_t 		*sigmsg = NULL;

	if (ftdm_queue_get_interrupt(sample_app_ftdm_init.event_msg_queue, ftdm_sangoma_isdn_int) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_CRIT, "Failed to get a event interrupt for app_freetdm_event_queue_handler thread !!\n");
		goto exit;
	}

	while (g_running) {
		ret_status = ftdm_interrupt_multiple_wait(ftdm_sangoma_isdn_int, 1, sleep);

		/* Find out why we returned from the interrupt queue */
		switch (ret_status) {
			case FTDM_SUCCESS:  /* There was a state change on the span */
				{
					/* It can possible that FreeTDM can pump lots of messages to queue, thus read all messages 
					 * untill queue is empty */
					while ((sigmsg = ftdm_queue_dequeue(sample_app_ftdm_init.event_msg_queue))) {
						ftdm_sigmsg_t *sigmsg_cpy = NULL;

						switch (sigmsg->event_id) {
							case FTDM_SIGEVENT_START:
								{
									sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
									memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
									ftdm_log(FTDM_LOG_DEBUG, "SIG_START event received. Starts io data operation for [s%dc%d]. \n",
											sigmsg->span_id, sigmsg->chan_id);

									/* Incoming Call. Proceed to answer the call and then
									   launch a media thread that will perform I/O on the channel */
									ftdm_channel_call_answer(sigmsg->channel);

									if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
										ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
												sigmsg->span_id, sigmsg->chan_id);
										break;
									}

									break;
								}

							case FTDM_SIGEVENT_UP:
								{
									sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
									memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
									ftdm_log(FTDM_LOG_DEBUG, "SIG_UP event received. Starts io data operation for [s%dc%d]. \n",
											sigmsg->span_id, sigmsg->chan_id);

									/* Outbound call has been answered, proceed to launch a meda thread that will
									   perform I/O on the channel */
									if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
										ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
												sigmsg->span_id, sigmsg->chan_id);
										break;
									}
									break;
								}

							case FTDM_SIGEVENT_RAW:
								{
									sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
									memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));

									if (sigmsg->raw.data) {
										sigmsg_cpy->raw.data = ftdm_malloc(sigmsg->raw.len);
										memcpy(sigmsg_cpy->raw.data, sigmsg->raw.data, sigmsg->raw.len);
									}

									ftdm_log(FTDM_LOG_DEBUG, "SIG_RAW event received\n");

									/* Handle Raw message received from lower layer */
									if (ftdm_thread_create_detached(tap_handle_raw_message, sigmsg_cpy) != FTDM_SUCCESS) {
										ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
												sigmsg->span_id, sigmsg->chan_id);
										break;
									}
									break;
								}

							default:
								break;
						}

						/* sample app owns the event buffer. It must be deallocated here
						   otherwise there is possibility of memory leak */
						ftdm_safe_free(sigmsg->raw.data);
						ftdm_safe_free(sigmsg);
					} /* while (dequeue) */
				}
				break;

			case FTDM_TIMEOUT:
				/* twiddle */
				break;

			case FTDM_FAIL:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned error!\n");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned with unknown code\n");
				break;
		} /* switch (ret_status) */
	} /* while */

exit:
	ftdm_log(FTDM_LOG_CRIT, "app_freetdm_event_queue_handler monitor thread stopping due to error.\n");
	return NULL;
}

#ifndef __WINDOWS__
/****************************/
static void signal_handler (int sig, siginfo_t *si, void *uc)
{
	ucontext_t* context = (ucontext_t*) uc;
	//print recv signal information
	printf("si_signo[%d], si_errno[%d] , si_code[%d],  si_pid[%d], si_uid[%d], si_status[%d], si_addr[%d], si_fd[%d] \n",
			si->si_signo, si->si_errno, si->si_code,
			si->si_pid, si->si_uid, si->si_status,
			si->si_addr, si->si_fd);

	switch (sig)
	{
		case SIGTERM:
			{
				printf("signal_handler: SIGTERM received \n");
				on_sigint(SIGTERM);
				break;
			}
		case SIGKILL:
			{
				printf("signal_handler: SIGKILL received \n");
				on_sigint(SIGKILL);
				break;
			}

		case SIGUSR1:
			{
				printf("signal_handler: SIGUSR1 received \n");
				on_sigusr(SIGUSR1);
				break;
			}
		case SIGUSR2:
			{
				printf("signal_handler: SIGUSR2 received \n");
				on_sigusr(SIGUSR2);
				break;
			}
		case SIGINT:
			{
				printf("signal_handler: SIGINT received \n");
				on_sigint(SIGINT);
				break;
			}

		default:
			printf("signal_handler: Unknown[%d] signal received \n", sig);
			break;
	}
}

void add_signal_handlers(int signal)
{
	struct sigaction sa;

	memset(&sa,0,sizeof(sa));
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = signal_handler;
	sigemptyset(&sa.sa_mask);
	sigaction(signal, &sa, NULL);
}
#endif

/**********************************************************************************
 * Fun   : main()
 *
 * Desc  : Configures all ports and spans as required by user and provide required
 * 	   platform for tapping calls 
 *
 * Ret   : err | FTDM_SUCCESS
**********************************************************************************/
int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	int err;
	int detected = 0;
	int configured = 0;
	sangoma_hw_config_reg_t reg;

	memset(&reg, 0, sizeof(reg));

#ifndef __WINDOWS__
	add_signal_handlers(SIGINT);
	add_signal_handlers(SIGTERM);
	add_signal_handlers(SIGKILL);
	add_signal_handlers(SIGUSR1);
	add_signal_handlers(SIGUSR2);
#if 0
	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGTERM, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);
#endif
#endif

	/* Initialize global FreeTDM init structure */
	memset(&sample_app_ftdm_init, 0, sizeof(sample_app_ftdm_init));

	/* Set basic debuggin parameters */
#ifdef SDK_DEBUGGING
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_DEBUG;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_ON_ASSERT;
#else
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_NEVER;
#endif

	if (FTDM_FAIL == app_open_log_file()) {
		fprintf(stderr, "Failed to open log file\n");
		return -1;
	}

	app_wp_config_input();

	/* set logger information if in case you dont have your own logger then please
	 * set sample_app_ftdm_init.logger = NULL to use default libsangoma freetdm logger */
	sample_app_ftdm_init.logger = app_logger;

	/* Initilize FreeTDM Framework 
	   Global Varialbes and FreeTDM Logger */
	err = sangoma_freetdm_init(&sample_app_ftdm_init);
	if (err) {
		/* close logging file */
		if (log_file_ptr) {
			fclose(log_file_ptr);
		}
		return err;
	}

	/* Initialize Sangoma Config SDK */
	reg.user_obj = NULL;
	reg.hwprobe_callback = sample_app_hwprobe_callback;
	reg.hwprobe_callback_post = sample_app_span_start_callback;
	err = sangoma_config_reg(&reg);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Sangoma Config SDK Initialization Error\n");
		exit_sample_app(max_wait);
	}

	/* Create a global FreeTDM event handling queue.
	 * FreeTDM event callback function is not allowed to exectue FreeTDM API commands.
	 * FreeTDM event callback fucntion will queue events, and the app event handler thread
	 * will consume and handle FreeTDM events.  
	 * The app event handler is allowed to run FreeTDM api commands directly */
	if (ftdm_thread_create_detached(app_freetdm_event_queue_handler, NULL) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create message queue handler thread. \n");
		exit_sample_app(max_wait);
	}

	/* Detect Sangoma Ports 
	   For each detected port run the sample_app_hwprobe_callback
	 */
	err = sangoma_config_hwprobe(&detected, &configured, SNG_ISDN_PRI_TAPPING_MODE);

	ftdm_log(FTDM_LOG_INFO, "Sangoma SDK Detected %d Configured %d\n", detected, configured);

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to detect sangoma hardware \n");

		/* We cannot exit here, as the thread is launched */
		exit_sample_app(max_wait);
	}
	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been created and started on each hw port */
	ftdm_sleep(1000);

	/* Enable Stats on all FTDM Span and Channels */
	sample_app_exec_io_stats("enable");

	/* Setting D channel wanpipe queue size */
	sample_app_set_rx_queuesize(900); 

	/* mark that initialization done */
	g_init=1;

	
	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	ftdm_log(FTDM_LOG_INFO, "Sangoma PRI/BRI TAPPING MODULE IN ACTION\n\n\n");
	while (tapping_main_menu()) {
		continue;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : exit_sample_app()
 *
 * Desc  : Stops all FTDM as well as application thread and deallocate the memory
 * 	   stop all the configured ports and exit from application.
 *
 * Ret   : NULL
**********************************************************************************/
int exit_sample_app(int max_wait)
{

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		ftdm_log(FTDM_LOG_INFO, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	if (g_init) {
		/* Stop FreeTDM Freamwork and all Spans */
		ftdm_global_destroy();
	}

	/* Stop all Sangoma hw ports */
	sangoma_stop_all_wanpipe();

	ftdm_sleep(1000);

	/* close logging file */
	if (log_file_ptr) {
		fclose(log_file_ptr);
	}
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : map_tapping_spans()
 *
 * Desc  : Detect the port number and configure span w.r.t tapping module, also
 * 	   peer span is selected depending up on the port and wanpipe number w.r.t
 * 	   that particular port 
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL | -1
**********************************************************************************/
int map_tapping_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t *span_conf)
{
	char peer_span_name[20];
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int port_num = sangoma_config_port_hwinfo_get_port_number(sangoma_port_obj);

	memset(span_conf, 0, sizeof(span_config_t));

	/* A102 has 2 ports - thus case 1 and 2 will match
	   A104 has 4 ports - thus case 1,2,3,4 will match
	   A108 has 8 ports */

	if ((port_num % 2) != 0) {
		wanpipe_num++;
	} else {
		wanpipe_num--;
	}

	sprintf(peer_span_name, "wp%d", wanpipe_num);

	switch (card_type) {
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
		{
			switch (port_num) {
			case 1:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 2:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 3:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 4:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 5:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 6:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 7:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			case 8:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				ftdm_set_string(span_conf->config.isdn.switch_type, "etsi");
				ftdm_set_string(span_conf->config.isdn.mixaudio_mode, "both");
				ftdm_set_string(span_conf->config.isdn.call_info_indication, "yes");
				//ftdm_set_string(span_conf->config.isdn.debug, "q921_raw");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "Invalid port id = %d\n", port_num);
				return FTDM_FAIL;
				break;
			}
		}
		break;

	case AFT_ADPTR_ISDN:
		{
			switch (port_num) {
			case 1:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				break;

			case 2:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				break;

			case 3:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				break;

			case 4:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				break;

			case 5:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "cpe");
				break;

			case 6:
				ftdm_set_string(span_conf->config.isdn.peer_span, peer_span_name);
				ftdm_set_string(span_conf->config.isdn.interface_name, "net");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "Invalid port id = %d\n", port_num);
				return -1;
				break;
			}
		}
		break;

	default:
		ftdm_log(FTDM_LOG_ERROR, "%s card (%d) only supports PRI/BRI Tapping\n", SNG_ISDN_TAP_CARD_TYPE(card_type),
				 card_type);
		return FTDM_FAIL;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : open_audio_file()
 *
 * Desc  : Open the the file to which raw data needs to  be written in order to tap
 * 	   the call.
 *
 * Ret   : -1 | fd
**********************************************************************************/
int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj)
{
	int fd = 0;
	char *path = (char *)obj;

	if (!path) {
		ftdm_log(FTDM_LOG_ERROR, "Cannot Open path as it is empty\n");
		return -1;
	}

	if ((fd = open(path, O_WRONLY | O_CREAT | O_TRUNC
					| FTDM_O_BINARY, S_IRUSR | S_IWUSR)) < 0 ) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to open file [%s], Error[%s]\n", path, strerror(errno));
		return -1;
	}
	ftdm_log(FTDM_LOG_DEBUG, "Start Tapping Call on channel %d:%d to [%s] having fd[%d]\n", span_id, chan_id, path, fd);
	return fd;
}

/**********************************************************************************
 * Fun   : app_wp_config_input()
 *
 * Desc  : Get wanpipe port configuration input 
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
ftdm_status_t app_wp_config_input()
{
	char input[128] = { 0 };
	int i = 0;

enter_media:
	/* get media type */
	fprintf(stdout, "Please Enter Media Type[T1=1,E1=2,FXOFXS=8] (int value) \n"); 
	if (NULL != fgets(input, 127, stdin)) {
		if ((input[0] != '\0') && (input[0] != '\n') && strlen(input)) {
			i = atoi(input);
			if (( i == 1) || (i==2) || (i==8)) {
				g_wp_config.media = i;
			} else {
				fprintf(stdout, "Invalid input plss enter any of [T1=1, E1=2, FXOFXS=8] \n");
				goto enter_media;
			}
		} else {
			fprintf(stdout, "Invalid input plss enter any of [T1=1, E1=2, FXOFXS=8]\n");
			goto enter_media;
		}
	} else {
		fprintf(stdout, "Invalid input plss enter any of [T1=1, E1=2, FXOFXS=8]\n");
		goto enter_media;
	}

	memset(input, 0 ,sizeof(input));

frame_mode:
	/* get frame mode  */
	fprintf(stdout, "Please Enter Framing Mode [ESF=1, D4=2, CRC4=4, NCRC4=5] (int value) \n"); 
	if (NULL != fgets(input, 127, stdin)) {
		if ((input[0] != '\0') && (input[0] != '\n') && strlen(input)) {
			i = atoi(input);
			if ((i==2) || (i == 1) || ( i == 4) || (i==5) ) {
				g_wp_config.frame_mode = i;
			} else {
				fprintf(stdout, "Invalid input plss enter any of [ESF=1, D4=2, CRC4=4, NCRC4=5] (int value) \n"); 
				goto frame_mode;
			}
		} else {
			fprintf(stdout, "Invalid input plss enter any of [ESF=1, D4=2, CRC4=4, NCRC4=5] (int value) \n"); 
			goto frame_mode;
		}
	} else {
		fprintf(stdout, "Invalid input plss enter any of [ESF=1, D4=2, CRC4=4, NCRC4=5] (int value) \n"); 
		goto frame_mode;
	}

	memset(input, 0 ,sizeof(input));
line_code:

	/* get line mode  */
	fprintf(stdout, "Please Enter Line Code [AMI=1, B8ZS=2, HDB3=3, B3ZS=4] (int value) \n"); 
	if (NULL != fgets(input, 127, stdin)) {
		if ((input[0] != '\0') && (input[0] != '\n') && strlen(input)) {
			i = atoi(input);
			if ((i==1) || (i == 2) || ( i == 3) || (i==4) ) {
				g_wp_config.line_code = i;
			} else {
				fprintf(stdout, "Invalid input plss enter any one of Line code [AMI=1, B8ZS=2, HDB3=3, B3ZS=4] (int value) \n"); 
				goto line_code;
			}
		} else {
			fprintf(stdout, "Invalid input plss enter any one of Line code [AMI=1, B8ZS=2, HDB3=3, B3ZS=4] (int value) \n"); 
			goto line_code;
		}
	} else {
		fprintf(stdout, "Invalid input plss enter any one of Line code [AMI=1, B8ZS=2, HDB3=3, B3ZS=4] (int value) \n"); 
		goto line_code;
	}

	memset(input, 0 ,sizeof(input));

get_impedence:
	/* get high impedence true/false mode  */
	fprintf(stdout, "Please Enter High Impedence value [OFF=0, ON=1] (int value) \n"); 
	if (NULL != fgets(input, 127, stdin)) {
		if ((input[0] != '\0') && (input[0] != '\n') && strlen(input)) {
			i = atoi(input);
			if ((i==0) || (i == 1)) {
				g_wp_config.high_impedance_mode = i;
			} else {
				fprintf(stdout, "Invalid input plss enter High Impedence value [OFF=0, ON=1] (int value) \n"); 
				goto get_impedence;
			}
		} else {
			fprintf(stdout, "Invalid input plss enter High Impedence value [OFF=0, ON=1] (int value) \n"); 
			goto get_impedence;
		}
	} else {
		fprintf(stdout, "Invalid input plss enter High Impedence value [OFF=0, ON=1] (int value) \n"); 
		goto get_impedence;
	}

	fprintf(stdout, "Wanpipe configuration: Media[%d], FrameMode[%d], LineMode[%d], HighImpedence[%d] \n", 
			g_wp_config.media, g_wp_config.frame_mode, g_wp_config.line_code, g_wp_config.high_impedance_mode);  

	g_wp_config.wp_config_pres=1;

	return FTDM_SUCCESS;
}

/******************************* Applications Logging module *************************************/

/**********************************************************************************
 * Fun   : app_open_log_file()
 *
 * Desc  : Get log file path and name from user else use default and open same in
 * 	   append mode so that logger can write in to it
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
ftdm_status_t app_open_log_file()
{
	char file_path[MAX_FILE_PATH_SIZE] = { 0 };

	/* get log file path and name and open the same in order to direct logs in to it*/
	fprintf(stdout, "Please Enter the file name with path in which SDK logs needs to be written: ");
	if (NULL != fgets(file_path, MAX_FILE_PATH_SIZE, stdin)) {
		file_path[strlen(file_path) - 1] = '\0';
		if (strlen(file_path)) {
			fprintf(stdout, "SDK log file directory: %s\n", file_path);
		} else {
			strcpy(file_path, "logs/sdk.log");
			fprintf(stdout, "Creating default i.e. %s file in current directory\n", file_path);
		}
	} else {
		strcpy(file_path, "logs/sdk.log");
		fprintf(stdout, "Creating default i.e. %s file in current directory\n", file_path);
	}

	fprintf(stdout, "Please press ENTER to continue...\n");
	getchar();

	if (file_path != NULL) {
		/* open file to write Debug prints i.e. logging in to a file */
		if ((log_file_ptr = fopen(file_path, "a")) == NULL) {
			fprintf(stderr, "Failed to open %s file\n", file_path);
			return FTDM_FAIL;
		}
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : app_logger()
 *
 * Desc  : This is application logger and mainly print logs on screen & file
 *
 * Ret   : void
**********************************************************************************/
void app_logger(const char *file, const char *func, int line, int level, const char *fmt, ...)
{
	char data[1024] = { 0 };
	char curr_time[80] = { 0 };
	va_list ap;
	const char *filep = (file ? sangoma_freetdm_cut_path(file) : "");

	if (level < 0 || level > 7) {
		level = 7;
	}

	va_start(ap, fmt);

	vsnprintf(data, sizeof(data), fmt, ap);

	/* get current date and time */
	sangoma_get_timestamp_str(curr_time);

	/* Avoid dumping everything to shell ..let that dump to a log file*/
	/* fprintf(stderr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data); */

	/* Dump only critical errors logging */
	if ((!strcasecmp("EMERG", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("CRIT", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("ERROR", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("WARNING", SANGOMA_FTDM_LEVEL_NAMES[level]))||
			(!strcasecmp("NOTICE", SANGOMA_FTDM_LEVEL_NAMES[level]))) {
		fprintf(stderr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);
	}

	/* dump logs in to file */
	if (log_file_ptr) {
		fprintf(log_file_ptr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);
	}

	va_end(ap);
}
